#ifndef CTX_H
#  define CTX_H 1

/********************

  Class context stores information about current scope.
  There could be following contexts:

    imported package context: (preexisted)
      contains classes only

    global context: (1 pass)
      contains classes only

    Class context (1 pass)
      Fields + Methods only: 

    Method block context (2nd pass)
      Generated on the fly during the second pass

****/
class Expression;
class FieldCtx;
class ClassCtx; 
class PackageCtx; 
class MethodCtx; 

void ReadBinaryClass(const char* name, ClassCtx* ctx);


struct Context : public QuickNew {
  static Context* current;
protected:

  struct CtxItem : public QuickNew {
    Identifier name;
    Context * ctx;
    struct CtxItem * nextInOwner;

    CtxItem(Identifier n, Context* c):name(n),ctx(c){}
  };
  mutable List<CtxItem> items;


  Context * findName(Identifier name) const { 
    for(CtxItem * p = items.first; p ; p = p->nextInOwner) 
      if(name == p->name) return p->ctx;
    return 0;
  }

  virtual bool replaceable() { return false; }

  void addName(Identifier name, Context* ctx) const {
    if(findName(name)) ASSERT(false, ER_DUPLICATED_NAME, name);
    items.add(new CtxItem(name, ctx) );
  }

  void addNameDontCheck(Identifier name, Context* ctx) {
    items.add(new CtxItem(name, ctx) );
  }
  
  bool findAndReplace(Identifier name,  Context* replace) {
    for(CtxItem * p = items.first; p ; p = p->nextInOwner) 
      if(name == p->name) {
        assert( p->ctx->replaceable() );
        p->ctx = replace; return true;
      }
    return false;    
  }

public:
  const Context * prev_ctx; 

  void setParentCtx(Context* parent) {prev_ctx = parent;}
  enum CtxType {PACKAGE, CLASS, FIELD, METHOD, LOCALVAR, BLOCK, ERROR, GLOBAL};
  static Context * errorCtx;

  const CtxType type;
  const Identifier fullName;
  int attributes;  

  Context():prev_ctx(0),type(ERROR),fullName(Identifier::crazy),attributes(0){}
  Context(const Context* parent, CtxType t, Identifier name=Identifier::crazy)
    :prev_ctx(parent),type(t),fullName(name), attributes(0){}

  Identifier combine(Identifier name1, Identifier name2) {
    if(name1.len() == 0) return name2;
    char buf[1024];
    char* a = name1.copyTo(buf);
    *a = '/';
    name2.copyTo(a+1);
    return Identifier(buf);
  }

  virtual Context* localFind(Identifier name) const { return findName(name); }

  Context * safeLocalFind(Identifier name) const {
    Context * r = localFind(name);
    ASSERT(r, ER_NOT_FOUND, name);
    return r;
  }

  virtual Context* selfCheck() const { return const_cast<Context*>(this); }

  virtual void finalize(){}

  virtual bool isInterface() { ASSERT(0, isInterface-not-applicable, fullName); return 0; }

  virtual bool isStatic() const 
    { ASSERT(false, ER_NOT_NONAPPLICABLE, fullName); return 0; }

  virtual bool isLocalVar() const 
    { ASSERT(false, ER_NOT_LOCAL_VAR, fullName); return 0; }

  virtual void resolveDot(Identifier) { ASSERT(false, ER_DOT_NONAPPLICABLE, fullName); }

  virtual void addFieldWithSig(int attr, Identifier name, Identifier typeSig) 
    { ASSERT(false, ER_NONAPPLICABLE, fullName); }

  virtual void addField(int attr, Identifier name, Type* type) 
    { ASSERT(false, ER_addField_NONAPPLICABLE, fullName); }

  //  virtual Identifier getTypeSig() const // for ClassCtx only
  //    { ASSERT(false, ER_getTypeSig_NONAPPLICABLE, fullName); return Identifier::crazy; }
 
  virtual Type* getType() const // for ClassCtx and FieldCtx
    { ASSERT(false, ER_getType_NONAPPLICABLE, fullName); return Type::Error; }
 
  virtual void addAnotherType(int, Type*) { // method overloading
    ASSERT(false, ER_DUPLICATED_NAME, fullName); }

  Context* find(Identifier name) const {   
    Context* result = localFind(name); 
    if(result) return result;
    //  cerr << "Context::find fall to ";
    //    if(prev_ctx) cerr << prev_ctx->fullName << endl;
    if(prev_ctx) return prev_ctx -> find(name);
    ASSERT(false, ER_NAME_NOT_FOUND, name);
    return errorCtx;
  }
  
  void indent(ostream& o, int level) { 
    for(int i = 0; i < level; ++i) cout << ' '; }

  void print(ostream& o, int level = 0) {
    cout << type << " ''" << fullName << "''" << endl;
    for(CtxItem* p = items.first; p != 0; p = p->nextInOwner) {
      indent(o, level+2); o << p->name << ' ';
      p->ctx->print(o, 2 + level);
    }
  }
};
Context* Context::current;

Context* Context::errorCtx     = new Context(0, Context::ERROR);

ClassCtx * asClassCtx(Context* ctx) {
  ASSERT(ctx->type == Context::CLASS, ER_CLASS_NAME_EXPECTED, ctx->fullName);
  return (ClassCtx*)ctx; // dynamic_class :(
}
MethodCtx * asMethodCtx(Context* ctx) {
  ASSERT(ctx->type == Context::METHOD, ER_METHOD_NAME_EXPECTED, ctx->fullName);
  return (MethodCtx*)ctx; // dynamic_class :(
}
FieldCtx * asFieldCtx(Context* ctx) {
  ASSERT(ctx->type == Context::FIELD, ER_VAR_OR_FIELD_EXPECTED, ctx->fullName);
  return (FieldCtx*)ctx; // dynamic_class :(
}
PackageCtx * asPackageCtx(Context* ctx) {
  ASSERT(ctx->type == Context::PACKAGE, ER_PACKAGE_NAME_EXPECTED, ctx->fullName);
  return (PackageCtx*)ctx; // dynamic_class :(
}

class FieldCtx : public Context { 
  u2 ref;
public:
  Type* const fieldType; 
  const Identifier typeSig;

  virtual bool isLocalVar() const { return attributes < 0; }

  virtual bool isStatic() const {
    ASSERT(!isLocalVar(), ER_THIS_IS_A_LOCALVAR, fullName);
    return attributes & M_STATIC;
  }

  FieldCtx(Context* parent, Identifier name, Type* type, int attr):
    Context(parent, FIELD, name), ref(0), typeSig( type->typeSig ), fieldType(type)
  { attributes = attr; }

  virtual Context* localFind(Identifier name) const 
    { ASSERT(false, ER_CANNOT_SEARCH_INSIDE_VAR, fullName); return 0; }

  int getSlot() {
    assert(attributes < 0);
    return -attributes - 1;
  }
  u2 getRef() {
    if(!ref) {
      u2 nametype = cpool.addNameAndType(fullName.no(), fieldType->typeSig.no());
      //      cerr << "GET_REF(" << nametype << ' ';
      ref = cpool.addFieldRef
	(CONSTANT_Fieldref, prev_ctx->getType()->classNo(), nametype);
      //      cerr << ref << ')';
    }
    return ref;
  }

  void issueLoad() {
    if(isLocalVar())
      code.local(op_iload, getSlot());
    else {
      if(isStatic())
	code.add(op_getstatic, 0, fieldType->stackSize());
      else 
	code.add(op_getfield,  1, fieldType->stackSize());

      code.ref(getRef());
    }
  }

  void issueStore() {
    if(isLocalVar())
      code.local(op_istore, getSlot());
    else {
      if(isStatic())
	code.add(op_putstatic, fieldType->stackSize(), 0);
      else 
	code.add(op_putfield,  1 + fieldType->stackSize(), 0);

      code.ref(getRef());
    }
  }

  void issueDeclaration() {
    code.beginField(attributes, fullName.no(), typeSig.no());
    code.add2(0); // does not handle ConstantValue yet
  }

};

class Prototype : public QuickNew {
  mutable MethodType * type;
  u2 ref;
  Context* ctx;
public:
  const Identifier typeSig;
  Prototype* nextInOwner;
  const int attributes;
  
  MethodType* getType() const {
    if(!type) type = asMethodType( Type::typeFromSig(typeSig) );
    return type;
  }
  u2 getRef() {
    if(!ref) {
      u1 refType = ctx->isInterface()?CONSTANT_IMethodref:CONSTANT_Methodref;
      u2 nametype = cpool.addNameAndType(ctx->fullName.no(), type->typeSig.no());
      ref = cpool.addFieldRef
	(refType, ctx->prev_ctx->getType()->classNo(), nametype);
    }
    return ref;
  }
  virtual bool isStatic() const { return attributes & M_STATIC;  }
  Prototype(int attr, Context* Ctx, Type* typ):ref(0),ctx(Ctx),
    type((MethodType*)typ),typeSig(typ->typeSig),attributes(attr) {assert(typ->isMethod());}
  virtual void qprint(ostream& o) const { 
    o << getType()->typeSig;
    if(isStatic()) o << " static ";
  }
};

class MethodCtx : public Context {
  List<Prototype> prototypes;
public:
  virtual bool isStatic() const { return attributes & M_STATIC;  }
  MethodCtx(Context* parent, Identifier name, Type* type, int attr):
    Context(parent, METHOD, name)
  {
    attributes = attr;
    prototypes.add(new Prototype(attr, this, type));
  }
  virtual bool isInterface() {
    assert(prev_ctx);
    return prev_ctx->attributes & M_INTERFACE;
  }
  void printPrototypes() {
    if(prototypes.count == 1) {
      cerr << "The only available prototype is " 
	   << fullName << prototypes.first << endl;
    } else {
      cerr << "Available prototypes are: " << endl;
      for(Prototype* p = prototypes.first; p; p = p->nextInOwner) {
        cerr << "  " << fullName << p << endl;
      }
    }
  }
  void addAnotherType(int attr, Type* type) {
    prototypes.add(new Prototype(attr, this, type)); // check for duplicates
    if( (attr&M_STATIC) != (attributes&M_STATIC) ) {
      cerr << endl;
      srcPos(cerr) << "jc will not handle inconsistent statics" << endl;
      printPrototypes();
      abort();
    }
  }
  Prototype * matchPrototype(Expression*,int); // defined in exprs.hh

};

class BlockCtx: public Context {
  int nLocals;
public:
  BlockCtx(int attr ):Context(Context::current, BLOCK) { 
    Context::current = this;
    attributes = attr;
  }
  BlockCtx():Context(Context::current, BLOCK) { 
    attributes = Context::current->attributes;
    Context::current = this;
  }
  //  BlockCtx(int attr, const Context* parent):Context(parent, BLOCK)
  //    { attributes = attr; } 
  virtual void finalize(){ 
    Context::current = const_cast<Context*>(prev_ctx);
    code.removeLocals(nLocals); 
  }
  virtual Context* localFind(Identifier name) const {
    Context* result = Context::localFind(name);
    //  cerr << "BlockCtx: localFind(" << name << ") " << (result?"found":"not found") << endl;
    return result;
  }
  virtual bool isStatic() const { return attributes & M_STATIC;  }  
  virtual void addField(int attr, Identifier name, Type* type) {
    //    cerr << "ADD FIELD: " << name << " " << typeSig << endl;
    nLocals += type->stackSize();
    addName(name, new FieldCtx(this, name, type, -1-code.addLocal(type)) );
  }
};

class ClassCtx : public Context {
  Type * super_class;
  int fields_count;
  int methods_count;
public:
  enum Enum {Mentioned, OnTheDisk, Ambiguous, Source, Loaded};
  mutable Enum state;

  static Type* NewClass(const ClassCtx* clazz) {
    if(clazz->self) return clazz->self;
    return clazz->self = Type::NewClass(clazz->fullName, clazz);
  }

  ClassCtx(const Context* parent, const Context * pkg, Identifier name, Enum st):
    Context(parent, CLASS, combine(pkg->fullName, name)),state(st),
    super_class(0),self(0) {}  

  bool canReach(Type * super) {
    if(super_class == super) return true;
    if(super_class == 0) return false;
    return super_class->getClassCtx()->canReach(super);
  }
    
  TypeList interfaces;
  int access_flags;
  mutable Type * self;

  virtual Type* getType() const {
    return NewClass(this);
  }

  virtual Context* localFind(Identifier name) const { 
    //  cerr << "ClassCtx_" << fullName << "::localFind(" << name << ')' << endl;
    (void)selfCheck();
    Context* ctx = findName(name);
    if(ctx) return ctx;
    if(super_class) ctx = super_class->getClassCtx()->localFind(name);
    if(ctx) return ctx;
    //    for(TypeListItem* t = interfaces.first; t != 0; t = t -> nextInOwner) {
    //     ctx = t->type->getClassCtx()->localFind(name);
    //     if(ctx) return ctx;
    //    }
    return 0;
  }

  virtual bool replaceable() { return (state == Ambiguous); }
  /*
  virtual Identifier getTypeSig() const {
    return NewClass(this)->typeSig;
  }
  */
  virtual Context* selfCheck() const { 
    switch(state) {
    case Mentioned: 
      ASSERT(false, ER_UNDECLARED_CLASS, fullName); break;
    case OnTheDisk: 
      ReadBinaryClass(fullName.asString(), const_cast<ClassCtx*>(this) );
      state = Loaded; break;
    case Ambiguous: 
      ASSERT(false, ER_AMBIGUOUS_CLASS, fullName); break;
    }
    return const_cast<ClassCtx*>(this);
  }

  void setSuperAndFlags(Identifier super, int flags) {
    super_class = Type::NewClass( super );
    access_flags = flags;
  }      
  void setSuperAndFlags(int flags) { // only for java/lang/Object
    super_class = 0; 
    access_flags = flags;
  }      
  void addInterface(Identifier name) {
    interfaces.add( Type::NewClass( name ) );
  }
  virtual void addField(int attr, Identifier name, Type* type) {
    Context* ctx = findName(name);
    //    cerr << "[ADD " << name << " " << typeSig << "]" << endl;
    if( ctx ) {
      ++methods_count;
      ctx -> addAnotherType(attr, type);
      return;
    }
    if( type->isMethod() ) {
      addName(name, new MethodCtx(this, name, type, attr) );
      ++methods_count;
    } else { 
      addName(name, new FieldCtx(this, name, type, attr));
      ++fields_count;
    }
  }

  void issueDeclarations() {
    code.add2(fields_count, "fields_count");
    for(CtxItem* f = items.first; f; f = f->nextInOwner)
      if(f->ctx->type == FIELD) ((FieldCtx*)(f->ctx))->issueDeclaration();
  }
};

class BinaryClassCtx : public ClassCtx {
public:
  BinaryClassCtx(const Context * pkg, Identifier name, bool read = true)
    :ClassCtx(0, pkg, name, read?Loaded:OnTheDisk) 
  {
    if(read) ReadBinaryClass(fullName.asString(), this);
  }
};

class PackageCtx : public Context {
  PackageCtx(Identifier name):Context(0, PACKAGE, name){}
  static Hash<Identifier, Context*, PKGS_HASH_SIZE> hash;
public:
  static Context* NewPackageCtx(Identifier name) {
    Context** pkg = &(hash.get(name).data);
    if(*pkg == 0) *pkg = new PackageCtx(name);
    return *pkg;
  }

  virtual Context* localFind(Identifier name) const { 
    //  cerr << "PackageCtx_" << fullName << "::localFind(" << name << ')' << endl;
    Context* result = findName(name);
    if(result) return result; // we already seen this name //
    if(exists( name.asString() )) result = NewPackageCtx(Identifier(fullName,'/',name));
    else result = new BinaryClassCtx(this, name);
    addName(name, result);
    return result; 
  }  
};
Hash<Identifier, Context*, PKGS_HASH_SIZE> PackageCtx::hash;

class GlobalCtx : public Context {
  Identifier packageName;
  char nameBuf[1024];
  int  nameLen;
  int mentionCount;
public:
  void setPackageName(Identifier name) {
    packageName = name;
    name.copyTo(nameBuf);
    nameLen = name.len();
    setParentCtx(PackageCtx::NewPackageCtx(name));
  }

  virtual Context* localFind(Identifier name) const { 
    //  cerr << "GlobalCtx::localFind(" << name << ')' << endl;
    Context* result = findName(name);
    if(result) return result->selfCheck();
    return 0; 
  }

  void importClass(Context* pkg, Identifier clazzName) {
    cout << "import " << clazzName << " from " << pkg->fullName << endl;
    Context* clazz = pkg->find(clazzName);
    ASSERT(clazz->type == CLASS, ER_IMPORT_CLASS_NOT_FOUND, clazzName);
    addName(clazzName, clazz);
  }
  void importPackage(Context* ctx) {
    assert(ctx->type == PACKAGE);
    PackageCtx* pkg = (PackageCtx*)(ctx);
    DIR *dirp;
    struct dirent *direntp;

    dirp = opendir(pkg->fullName.asString());
    ASSERT(dirp, ER_PACKAGE_NOT_FOUND, pkg->fullName);    
    while( (direntp = readdir( dirp )) != NULL ) {
      const char* name = direntp->d_name;
      int len = strlen(name);
      if(len > 6 && strncmp(".class", name + len - 6, 6) == 0 ) {
        Identifier id(name, len - 6);
        ClassCtx* ctx = (ClassCtx*)findName(id); // Only classes stored on the global level
        if(ctx) ctx -> state = ClassCtx::Ambiguous;
	else addNameDontCheck(id, new BinaryClassCtx(pkg, id, false));
      }
    }
    closedir(dirp);
  }
  GlobalCtx():nameLen(0),Context(0, GLOBAL, Identifier::nullstr){
    Context::current = this;
    cout << "Loading java.lang.* ..." << flush;
    importPackage(PackageCtx::NewPackageCtx(Identifier("java/lang")));
    cout << "done" << endl;
  }
  Context*  mention(Identifier name) {
    //    cerr << "mention(" << name << ")";
    Context* clazz = findName(name);
    if(clazz) return clazz;
    /*    if(prev_ctx) {
      clazz = prev_ctx->localFind(name);
      if(clazz->type == CLASS) return clazz;   
    }*/
    clazz = new ClassCtx(this, this, name, ClassCtx::Mentioned);
    addName(name, clazz);
    ++mentionCount;
    return clazz;
  }

  Context* _findClassByQualifiedName(const char * name, int len) {
    int i = len;
    while( (--i >= 0) && (name[i] != '/'));
    if(i < 0) return mention(Identifier(name, len));
    Identifier pkgName(name, i), clazzName(name + i + 1, len - i - 1);
    //    cerr << pkgName << '.' << clazzName << " Global Package = " << packageName << endl;
    if(pkgName == packageName) return mention(clazzName);
    Context* pkg = PackageCtx::NewPackageCtx(pkgName);   
    return pkg->find(clazzName);
  } 

  ClassCtx* findClassByQualifiedName(const char * name, int len) {
    Context* ctx = _findClassByQualifiedName(name, len);
    assert(ctx->type == CLASS);
    return (ClassCtx*)ctx;
  }

  void checkMentionedTypes() {
    if(Identifier("(c)Gorik").no() == 0) cerr << "Strange" << endl;
    if(mentionCount == 0) return;
    mentionCount = 0;
    for(CtxItem * p = items.first; p ; p = p->nextInOwner) 
      if(p->ctx->type == CLASS) {
        ClassCtx * cl = (ClassCtx*)(p->ctx);
        if(cl->state == ClassCtx::Mentioned) {
	  srcPos(cerr) << "ERROR: name ``" << cl->fullName 
		       << "'' mentioned, but never defined" << endl;
	  ++mentionCount;
	}
      }
    if(mentionCount) abort();
    else srcPos(cerr) << "complain: mentionCount != 0. Strange..." << endl;
  }

  ClassCtx* findClassByQualifiedName(Identifier name) 
    { return findClassByQualifiedName(name.asString(), name.len()); }

  ClassCtx* findClassBySignature(Identifier typeSig) 
    { return findClassByQualifiedName(typeSig.asString()+1, typeSig.len()-2); }

  ClassCtx* defineClass(Identifier name) {
    ClassCtx * clazz =  asClassCtx( mention(name) );
    if(clazz->state == ClassCtx::Ambiguous || clazz->state == ClassCtx::OnTheDisk) {
      clazz = new ClassCtx(this, this, name, ClassCtx::Source);
      findAndReplace(name, clazz);
      return clazz;
    }
    ASSERT(clazz->state == ClassCtx::Mentioned,
      ER_REDEFINITION_OF_THE_CLASS, name);
    --mentionCount;
    return clazz;
  }

};

GlobalCtx globalCtx;
ClassCtx* globalCtx_findClass(Identifier clazzName) {
  //cerr << "gCtx(" << clazzName << ")" << endl;
  return globalCtx.findClassByQualifiedName(clazzName);
}

#endif
