/************************************************
  Team: Jon Gregory, Gor Nishanov
  Project #3
    for RPI course 66-648 Compiler Design Theory

  stmts.h:
    clases representing java statements

    */

struct Label : public QuickNew {
  Identifier name;
  Label* nextInOwner;
  
  Label(Identifier nam):name(nam),nextInOwner(0){}
};

class CaseStatement;
class LoopStatement;

class Statement : public Node {
  List<Label> labels;

  Statement* prev_loop;
  static Statement* currentLoop;

  friend LoopStatement;
public:

  bool hasLabel(Identifier name) {
    for(Label* l = labels.first; l; l = l->nextInOwner) 
      if(name == l->name) return true;
    return false;
  }

  Statement* findLoop(Identifier name) {
    if(!(name == Identifier::nullstr)) {
      for(Statement* l = currentLoop; l; l = l->prev_loop) {
	if(l->hasLabel(name)) return l;
      }
      cerr << "Warning: loop label " << name << " not found" << endl;
    }
    ASSERT(currentLoop, ER_NO_CURRENT_LOOP, this);
    return currentLoop;
  }

  explicit Statement(int op):Node(op),nextInOwner(0),prev_loop(0){}
  
  void    addLabel(Label* lab){labels.add(lab); }
  void    addLabel(Identifier id){addLabel(new Label(id) ); }
  
  void printIndent(ostream& o, int indent) const {
    for (int i = 0 ; i < indent ; i++) o << "  ";	}
  
  virtual void print2(ostream& o, int) const {
    Label* l = labels.first;
    while(l != 0) {
      o << l->name << ": ";
      l = l -> nextInOwner;
    }
  }
 
  virtual void issueCode() { 
    cerr << "issueCode: NOT IMPLEMENTED" << this << endl; exit(0);  }
  virtual void issueLoopCode() { 
    cerr << "issueLoopCode: NOT IMPLEMENTED" << this << endl; exit(0);  }
 
  virtual bool isSuper() { return false; }

  virtual int continueAddr() 
    {cerr << "continue: NOT IMPLEMENTED " << this << endl; return 0; }

  virtual int breakAddr() 
    {cerr << "break: NOT IMPLEMENTED " << this << endl; return 0; }

  virtual void addCaseLabel(CaseStatement*, bool) 
    {ASSERT(false, ER_CASE_NOT_INSIDE_SWITCH, this);}

  virtual void print(ostream& o) const {print2(o, (int)0);}

  virtual void completeDeclaration(Context* ctx) {
    ASSERT(false, ER_NOT_A_DECLARATION, this);
  }

  static void stmtPrint(ostream& o, Statement* st, int indent) {
    if(st == null) o << "<empty>";
    else st->print2(o, indent);
  }
  
  Statement * glue(Statement* b) {
    nextInOwner = b;
    return this;
  }
  Statement * nextInOwner;
};
Statement* Statement::currentLoop = 0;
typedef List<Statement> StmtList;

class LoopStatement : public Statement {
protected:
  int  contAddr;
  int   brkAddr;
public:
  explicit LoopStatement(int op):contAddr(0),brkAddr(0),Statement(op){}  

  virtual int continueAddr() { 
    ASSERT(contAddr, ER_NOT_ALLOWED, this);
    return contAddr; 
  }

  virtual int    breakAddr() { 
    ASSERT(brkAddr, ER_NOT_ALLOWED, this);
    return brkAddr; 
  }

  virtual void issueCode() {
    prev_loop = currentLoop;
    currentLoop = this;
    issueLoopCode();
    currentLoop = prev_loop;
    if(brkAddr) code.patch_branch(brkAddr);
  }
};

class BreakStatement : public Statement {
  Identifier lbl;
public:
  BreakStatement(Identifier label):Statement(BREAK), lbl(label) {}
  
  virtual void issueCode() {
    code.branch(op_goto, findLoop(lbl)->breakAddr());
  }
  
  virtual void print2(ostream& o, int indent) const {
    Statement::print2(o, indent);
    o << "break";
    if (!(lbl == Identifier::nullstr)) {
      o << ' ' << lbl; }
    o << ';';
  }
};

class ContinueStatement : public Statement {
  Identifier lbl;
public: ContinueStatement(Identifier label):Statement(CONTINUE),lbl(label){}

  virtual void issueCode() {
    code.branch(op_goto, findLoop(lbl)->continueAddr());
  }
  
  virtual void print2(ostream& o, int indent) const {
    Statement::print2(o, indent);
    o << "continue" << ' ' << lbl << ';';
  }
};

class CatchStatement : public Statement {
  Expression * throwExpr;
  Identifier id;
  Statement * body;
  Field * field;
public:
  CatchStatement(Expression* Expr, Identifier Id, Statement* Body):
    Statement(CATCH), throwExpr(Expr), id(Id), body(Body), field(0) {}
  
  virtual void print2(ostream& o, int indent) const {
    Statement::print2(o, indent);
    o << "catch (" << throwExpr;
    o << ' ' << id  << ") ";
    if (body != null) {	body->print2(o, indent);	}
    else o << "<empty>";
  }
};

class CompoundStatement : public Statement {
  Statement* stmts;
public:
  CompoundStatement(Statement* Stmts):Statement(STAT),stmts(Stmts){}

  void issueMethodCode(int modifiers, Parameter* p, bool checkSuper = false) {
    Context::current = new BlockCtx(modifiers);
    if((modifiers & M_STATIC) == 0) 
      Context::current->addField(0, Identifier::self, 
			   Context::current->prev_ctx->getType());

    for(; p; p = p->nextInOwner) {
      //      cerr << "adding param " << p->name << ' ' << p->type << endl;
      Context::current->addField(0, p->name, p->type);
    }

    if( stmts ) {
      stmts->issueCode();
      code.checkStack();
      if(checkSuper && stmts->isSuper()) {
	for(Statement* i = Assembler::pre_ctor; i; i = i -> nextInOwner) {
	  i->issueCode();
	  code.checkStack();
	}
      }
      for(Statement* s = stmts->nextInOwner; s; s = s->nextInOwner) {
	s->issueCode();
	code.checkStack();
      }
    }
    Context::current->finalize();
  }
  
  virtual void issueCode() {
    Context::current = new BlockCtx();
    for(Statement* s = stmts; s; s = s->nextInOwner) {
      s->issueCode();
      code.checkStack();
    }
    Context::current->finalize();
  }
  
  virtual void print2(ostream& o, int indent) const {
    Statement::print2(o, indent);
    cout << '{' << endl;
    Statement* s = stmts;
    while(s != 0) {
      printIndent(o, indent+1);
      s -> print2(o, indent + 1);
      o << endl;
      if(s->isSuper()) {
	for(Statement* i = Assembler::pre_ctor; i; i = i -> nextInOwner) {
	  printIndent(o, indent+1);
	  i -> print2(o, indent + 1);
	  o << endl;
	}
      }
      s = s->nextInOwner;
    }
    printIndent(o, indent);	o << '}';
  }
};

class DoStatement : public LoopStatement {
  Statement *body;
  Expression *cond;
  
public:
  DoStatement(Statement* Body, Expression* Cond):
    LoopStatement(DO), body(Body), cond(Cond) {}
  
  virtual void issueLoopCode() {
    int firstGoto = code.branch(op_goto, 0);
    brkAddr = code.branch(op_goto, 0);
    code.patch_branch(firstGoto);
    int whileAddr = code.currentAddress();
    if(body) body->issueCode();
    cond->codeBranchIfTrue(whileAddr);
  }
  
  virtual void print2(ostream& o, int indent) const {
    Statement::print2(o, indent); o << "do ";
    body -> print2(o, indent);
    o << " while " << *cond << ';';
  }
};

class EmptyStatement : public Statement {
public:
  EmptyStatement():Statement(VOID) {}
  virtual void issueCode() {}
  virtual void print2(ostream& o, int) const { o << ';' ; }
};

class DeclarationStatement : public Statement {
  int modifiers;
  Type * type;
  Expression * vars;

public:
  DeclarationStatement(int attr, Type *typ, Expression * arg):
    Statement(DECLARATION),modifiers(attr),type(typ),vars(arg){}

  virtual void completeDeclaration(Context* ctx); // defined in tree.h

  virtual void issueCode() {
    Expression * e = vars;
    while(e) {
      Expression * i = e->getDeclaringIdent();
      //      cerr << "declaring " << i->toIdent() << endl;
      Context::current->addField(modifiers, i->toIdent(), type->dims(i->getDims()) );
      if(e->op == ASSIGN) e->codeVoid();
      e = e->nextInOwner;
    }
  }

  virtual void print2(ostream& o, int indent) const {
    o << "declare "; printMod(o, modifiers);
    Statement::print2(o, indent);
    o << type << ' ';
    Expression* s = vars;
    while(s != 0) {
      if(s != vars) o << ", ";
      s -> print(o);
      s = s -> nextInOwner;
    }
    o << ';';
  }
};

class FinallyStatement : public Statement {
  Statement* body;
  Statement* finalBody;
  bool finallyFinishes; // does finalBody never return?
  
public:
  FinallyStatement(Statement* Body, Statement* finalbody):
    Statement(FINALLY),body(Body), finalBody(finalbody), finallyFinishes(0) {}

  virtual void issueCode() { //
    if(body) body->issueCode();
    if(finalBody) body->issueCode();
  }

  virtual void print2(ostream& o, int indent) const {
    Statement::print2(o, indent);
    o << "try ";
    stmtPrint(o, body, indent);
    o << "finally ";
    stmtPrint(o, finalBody, indent);
  }
};

class ExpressionStatement : public Statement {
  Expression* expr;
public:
  ExpressionStatement(Expression* e):Statement(EXPRESSION),expr(e) {}

  virtual bool isSuper() { return expr->isSuper(); }

  virtual void issueCode() { expr->codeVoid(); }
  
  virtual void print2(ostream& o, int indent) const {
    Statement::print2(o, indent);
    o << expr << ';';
  }
};

class ForStatement : public LoopStatement {
  Statement* init;
  Expression* cond;
  Expression* inc;
  Statement* body;

public:
  ForStatement(Statement* ini, Expression* c, Expression* incr,Statement* Body)
    :LoopStatement(FOR),init(ini),cond(c), inc(incr), body(Body) {}

  virtual void issueLoopCode() {
    Context::current = new BlockCtx();

    init->issueCode();

    int firstGoto = code.branch(op_goto, 0); // will patch it later //

    brkAddr = code.branch(op_goto, 0); // patched in LoopStatement //
    contAddr  = code.currentAddress();    
    if(inc) inc->codeVoid();

    code.patch_branch(firstGoto);
    if(cond) cond->codeBranchIfFalse( brkAddr );

    body->issueCode();

    code.branch(op_goto, contAddr);

    Context::current->finalize();
  }

  virtual void print2(ostream& o, int indent) const {
    Statement::print2(o, indent);
    o << "for (";
    if (init != null) {
       init -> print2(o, indent); o << ' ';
   } else {
     o << "; ";
   }
   if (cond != null) {
     cond->print(o);
     o << " ";
  }
  o << "; ";
  if (inc != null) {inc->print(o);}
     o << ") ";
     if (body != null) {
       body->print2(o, indent);
     } else {
       o << ';';
     }
  }
};


class IfStatement : public Statement {
  Expression* cond;
  Statement* ifTrue;
  Statement* ifFalse;
  
public:
  IfStatement(Expression* c, Statement* Then, Statement* Else):
    Statement(IF),cond(c),ifTrue(Then),ifFalse(Else) {}

  void issueCode() {
    int firstGoto = code.branch(op_goto, 0); // to patch later
    int elseAddr = code.currentAddress();
    if(ifFalse) ifFalse->issueCode();
    int brkAddr = code.branch(op_goto, 0); // to patch later 

    code.patch_branch(firstGoto); // firstGoto goes here
    cond->codeBranchIfFalse(elseAddr);
    if(ifTrue) ifTrue->issueCode();
    code.patch_branch(brkAddr);
  }
  
  virtual void print2(ostream& o, int indent) const {
    Statement::print2(o, indent);
    o << "if " << cond << ' ';
    ifTrue -> print2(o, indent);
    if (ifFalse != null) {
      o << " else ";
      ifFalse -> print2(o, indent);
    }
  }
};

class ReturnStatement : public Statement {
  Expression* expr;
public:
  ReturnStatement(Expression* e):Statement(RETURN),expr(e){}
  
  virtual void issueCode() {
    if (expr == null) {
      code.add(op_return);
      return;
    }
    expr->codeValue();
    code.add(op_ireturn + expr->type->offset());
  }

  virtual void print2(ostream& o, int indent) const {
    Statement::print2(o, indent);
    o << "return";
    if (expr != null) o << ' ' << expr;
    o << ';';
  }
};

class TryStatement : public Statement {
  Statement* body;
  Statement* args; // []
  
public:
  TryStatement(Statement* Body, Statement* arg):
    Statement(TRY), body(Body), args(arg){}

  virtual void issueCode() {
    if(body) body->issueCode();
  }
  
  virtual void print2(ostream& o, int indent) const {
    Statement::print2(o, indent);
    o << "try ";
    stmtPrint(o, body, indent);
    Statement *s = args;
    while(s != 0) {
      o << ' ' << s;
      s = s->nextInOwner;
    }
  }
};

class SwitchStatement;

class CaseStatement : public Statement {
  Expression* expr;
  
  CaseStatement* nextCase;
  int            addr;
  int            value;

  friend SwitchStatement;
  
public:

  virtual void issueCode() {
    ASSERT(currentLoop, ER_CASE_WITHOUT_SWITCH, this);
    addr = code.currentAddress();
    if(expr) {
      ASSERT(expr->op == INTVAL, ER_ONLY_CONST_INT_ALLOWED, this);
      value = ((IntExpression*)expr)->value;
      currentLoop->addCaseLabel(this, false);
    } else {
      currentLoop->addCaseLabel(this, true);
    }
  }

  CaseStatement(Expression* e):Statement(CASE),expr(e),nextCase(0) {}

  virtual void print2(ostream& o, int indent) const {
    Statement::print2(o, indent);
    if (expr == null) o << "default";
    else { o << "case "; safePrint(o, expr); }
    o << ':';
  }
};

class SwitchStatement : public LoopStatement {
  Expression* expr;
  Statement* body; //[];

  CaseStatement* first;
  CaseStatement* last;
  CaseStatement* defLabel;
  int count;
  
public:
  virtual void addCaseLabel(CaseStatement* l, bool def) {
    if(def) {
      ASSERT(!defLabel, ER_DEFAULT_ALREADY_DEFINED, this);
      defLabel = l;
      return;
    }
    if(count == 0) first = last = l;
    else last = (last->nextCase = l);
    ++count;
  } 

  SwitchStatement(Expression* e, Statement* stmts):
    LoopStatement(SWITCH),expr(e),body(stmts),count(0),defLabel(0) {}

  virtual void issueLoopCode() {
    int gotoAddr = code.branch(op_goto, 0); // will be patched
         brkAddr = code.branch(op_goto, 0);
    body->issueCode();
    code.branch(op_goto, brkAddr);

    code.patch_branch(gotoAddr);

    expr->codeValue();
    ASSERT(expr->type->inMask(TM_INTEGER), ER_ONLY_INTEGERS_ALLOWED, this);
    Expression::codeConversion(expr->type, Type::Int);

    int switch_start = code.currentAddress();
    code.add(op_lookupswitch);
    code.pad(4);
    int after_padding = code.currentAddress();
    if(defLabel) code.add4(defLabel->addr - switch_start);
    else code.add4(after_padding + 8 * (count+1) - switch_start);
    code.add4(count);
    for(int i = 0; i < count; ++i) {
      code.add4(first->value);
      code.add4(first->addr - switch_start);
      first = first->nextCase;
    }
  }
  
  virtual void print2(ostream& o, int indent) const {
    Statement::print2(o, indent);
    o << "switch (" << expr << ") {" << endl;
    if(body != 0) {
      printIndent(o, indent + 1);
      body -> print2(o, indent + 1); o << endl;
    }
    printIndent(o, indent);
    o << '}';
  }
};

class WhileStatement : public LoopStatement {
  Expression* cond;
  Statement* body;
public:
  WhileStatement(Expression* c, Statement* b):
    LoopStatement(WHILE),cond(c), body(b) {}

  virtual void issueLoopCode() {
     contAddr = code.branch(op_goto, 0); // to patch
      brkAddr = code.branch(op_goto, 0);
    int whileAddr = code.currentAddress();
    if(body) body->issueCode();
    code.patch_branch(contAddr);
    cond->codeBranchIfTrue(whileAddr);
  }
  
  virtual void print2(ostream& o, int indent) const {
    Statement::print2(o, indent);
    o << "while " << cond;
    if (body != null) {
      o << ' ';
      body -> print2(o, indent);
    } else {
      o << ';';
    }
  }
};

class SynchronizedStatement : public Statement {
  Expression* expr;
  Statement* body;
  
public:
  SynchronizedStatement(Expression* e, Statement* b):
    Statement(SYNCHRONIZED), expr(e), body(b) {}

  virtual void issueCode() {
    int slot = code.addLocal(expr->type);

    expr->codeValue();
    code.local(op_istore, slot);
    code.local(op_iload, slot);
    code.add(op_monitorenter);

    body->issueCode(); // correct implementation should
                    // handle possible exception here

    code.local(op_iload, slot);
    code.add(op_monitorexit);
    code.removeLocals(1);
  }
  
  virtual void print2(ostream& o, int indent) const {
    Statement::print2(o, indent);
    o << "synchronized " << expr << ' ';
    stmtPrint(o, body, indent);
  }
};

class ThrowStatement : public Statement {
  Expression* expr;
  
public:
  ThrowStatement(Expression* e):Statement(THROW),expr(e){}

  virtual void issueCode() {
    expr->codeValue();
    code.add(op_athrow);
  }

  virtual void print2(ostream& o, int indent) const {
    Statement::print2(o, indent);
    o << "throw " << expr << ';';
  }
};


