/************************************************
  Team: Jon Gregory, Gor Nishanov
  Project #3
    for RPI course 66-648 Compiler Design Theory

  tree.h:

    Here we have large classes representing
      classes, methods, packages and whole compiling environment

      **/

#include <iomanip.h>
#include <fstream.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <dirent.h>

class Field;

class Type;
class TypeListItem;
class Expression;
class Statement;

#include "consts.hh"
#include "opcodes.hh"
#include "cpool.hh"
#include "utils.hh"

#include "types.hh"
#include "asm.hh"
#include "ctx.hh"

#include "loader.hh"
#include "exprs.hh"
#include "stmts.hh"

class Field : public QuickNew {
public:
  Identifier name;
  int        modifiers;
  Type *     type;
  
  Field * nextInOwner;
};

class Method : public QuickNew {
  Identifier name;
  bool ctor;
public:
  int dims;
  TypeList throwz;
  int             modifiers;
  List<Parameter> params;
  CompoundStatement*    body;
  Type*           returnType;
  Type*           prototype;

  Method(Identifier Name, bool Ctor = false):
    name(Name),dims(0),ctor(Ctor),prototype(0){}

  void addParameter(Expression* name, Type* type) {
    params.add(new Parameter(
      name->toIdent(),
      type->dims(name->getDims() )
    ));
  }

  void issueCode() {

    code.beginMethod(modifiers, name.no(), prototype);
    body->issueMethodCode(modifiers, params.first, name == Identifier::ctor); 
    code.endMethod();

  }

  friend ostream& operator << (ostream& o, Method* m) {
    o << endl << "  "; 
    printMod(o, m->modifiers);
    o << m->name << m->prototype << ' ';
    if(m -> body != 0) m->body->print2(o, 1);
    return o;
  }

  void completeMethod(ClassCtx* ctx) {
    prototype = Type::NewMethod(returnType, params.first);
    ctx->addField(modifiers, name, prototype);
  }

  Method*  nextInOwner;
};

class Class : public QuickNew {
  int        modifiers;

public:
  const Identifier name;
  ClassCtx * ctx;
  Type * self;

  Class(Identifier nam, int mods):name(nam),modifiers(mods), 
    ctx( globalCtx.defineClass(name) ), ctors(0),
    self(Type::NewClass(ctx->fullName, ctx) )
  { }

  TypeList interfaces;
  List<Statement> fields;
  List<Method> methods;
  List<Statement> clinit;
  List<Statement> init;
  int ctors;

  Type * super;

  void completeClass() {
    ctx->setSuperAndFlags(super->getClassName(), modifiers);
    for(Statement* f = fields.first; f ; f = f -> nextInOwner) {
      f->completeDeclaration(ctx);
    }
    if(ctors == 0) { // we need to add default ctor
      Method * ctor = new Method(Identifier::ctor, true);
      ctor -> modifiers = M_PUBLIC;
      ctor -> returnType = Type::Void;
      ctor -> body = new CompoundStatement(
	new ExpressionStatement( new MethodExpression( new SuperExpression(super), &empty_exprs) )
      );
      methods.add( ctor );
    }
    if(clinit.count){ // create <clinit>
      Method * cinit = new Method(Identifier::clinit);
      cinit -> modifiers = M_STATIC;
      cinit -> returnType = Type::Void;
      cinit -> body = new CompoundStatement( clinit.first );
      methods.add( cinit );
    }
    for(Method* m = methods.first; m ; m = m -> nextInOwner) {
      m->completeMethod(ctx);
    }
    ctx->state = ClassCtx::Source;
  }

  void issueCode() {
    Context::current = ctx;
    Assembler::pre_ctor = init.first;

    code.beginClass(modifiers, self->classNo(), super->classNo());
    
    code.add2(interfaces.count, "interfaces_count");
    for(TypeListItem* t=interfaces.first; t; t=t->nextInOwner) code.add2(t->type->classNo());

    ctx->issueDeclarations(); // will print fields 

    code.add2(methods.count, "methods_count");
    for(Method* m = methods.first; m; m = m->nextInOwner) m->issueCode();

    code.add2(0, "attributes_count");

    code.endClass();
  }

  friend ostream& operator << (ostream& o, Class* c) {
    Assembler::pre_ctor = c->init.first;

    o << endl << "class " << c->name << " extends " 
      << c->super << " implements " << c->interfaces << " {";

    if(c->fields.count > 0) o << "\n  ";
    c->fields.print(o, "\n  ");
    o << c->methods;
    o << endl <<'}' << endl;
    return o;
  }

  Class * nextInOwner;
};

class Package : public QuickNew {
  Expression * name;
public:
  Package(Expression * nam): name(nam){
    // Add Proper Handling of package name
    globalCtx.setPackageName(Identifier::nullstr);
  }

  List<Expression> imports;
  List<Class>      classes;  

  void issueCode() {
    globalCtx.checkMentionedTypes();
    for(Class* cl = classes.first; cl; cl = cl->nextInOwner) cl->issueCode();
  }

  friend ostream& operator << (ostream& o, const Package* p) {
    o << "package " << p->name << ';'  << endl;
    if (p->imports.count > 0) 
      o << "import " << p->imports << ';' << endl;
    o << p->classes << endl << "---- end package ---" << endl;
    return o;
  };

  Package * nextInOwner;
};

class Environment : public QuickNew {
  Package * findPackage(const char * id) {
    return (Package*) 0; // Later we package search here
  }

public:
  CompoundStatement * empty_block;
  Environment():pkg(0),clazz(0),empty_block(new CompoundStatement(0)){
    //    cout << "Environment has been created" << endl;
  }
  Package * pkg; // currentPackage
  Class * clazz; // currentClass
  Method * method; // currentMethod

  void addMethod(Method * m);
  void classComplete();
  void addNewClass(Class * cl);

  //  Expression * arrayElem; // used in New Array

  void startArray(Type* type) {
    gen_indices.clear();
    gen_type = type;
  } 
  void startArray(Expression* name) {
    startArray(ClassCtx::NewClass(name->resolveClassName()));  }

  void addDim(Expression* dim) {
    gen_type = Type::NewArray( gen_type );
    if(dim) gen_indices.add(dim);
  }

  List<Expression> gen_indices; // used in generate array
  Type *           gen_type;
};

Environment env;

void Environment::addNewClass(Class * cl) {
  pkg->classes.add(cl);
  clazz = cl;
}
void Environment::classComplete() {
  clazz->completeClass();
  clazz = 0;
}
void Environment::addMethod(Method * m) {
  clazz -> methods.add(m);
  method = m;
}

void DeclarationStatement::completeDeclaration(Context* ctx) {
  Expression * e = vars;
  while(e) {
    Expression * i = e->getDeclaringIdent();
    if(e->op == ASSIGN) {
      if(modifiers & M_STATIC) env.clazz->clinit.add( new ExpressionStatement(e) );
      else env.clazz->init.add( new ExpressionStatement(e) );
      // !! need special processing for M_FINAL //
    }
    ctx->addField(modifiers, i->toIdent(), type->dims(i->getDims()) );
    e = e->nextInOwner;
  }
}

